from functools import wraps, partial
import inspect
from json import JSONEncoder
from threading import local as threadlocal
from typing import AnyStr, Tuple, Optional
import warnings
import copy
import logging
from datetime import datetime, timedelta
from urllib.parse import urlparse, urlunsplit

MAX_PAYLOAD_LENGTH = 128 * 1024
MAX_STRING_LENGTH = 1024


__all__ = []  # type: ignore


class SanitizingJSONEncoder(JSONEncoder):
    """
    A JSON encoder which handles filtering and conversion from JSON-
    incompatible types to strings.

    >>> import logging
    >>> from json import loads
    >>> logger = logging.getLogger(__name__)
    >>> encoder = SanitizingJSONEncoder(logger, keyword_filters=['bananas'])
    >>> items = loads(encoder.encode(FilterDict({'carrots': 4, 'bananas': 5})))
    >>> items['bananas']
    '[FILTERED]'
    >>> items['carrots']
    4
    """

    filtered_value = '[FILTERED]'
    recursive_value = '[RECURSIVE]'
    unencodeable_value = '[BADENCODING]'

    def __init__(self, logger: logging.Logger, keyword_filters=None, **kwargs):
        self.logger = logger
        self.filters = list(map(str.lower, keyword_filters or []))
        self.bytes_filters = [x.encode('utf-8') for x in self.filters]
        super(SanitizingJSONEncoder, self).__init__(**kwargs)

    def encode(self, obj):
        safe_obj = self._sanitize(obj, False)
        payload = super(SanitizingJSONEncoder, self).encode(safe_obj)
        if len(payload) > MAX_PAYLOAD_LENGTH:
            safe_obj = self._sanitize(safe_obj, True)
            return super(SanitizingJSONEncoder, self).encode(safe_obj)
        else:
            return payload

    def filter_string_values(self, obj, ignored=None, seen=None):
        """
        Remove any value from the dictionary which match the key filters
        """
        if not ignored:
            ignored = set()

        # Keep track of nested objects to avoid having references garbage
        # collected (which would cause id reuse and false positive recursion
        if seen is None:
            seen = []

        if type(ignored) is list:
            ignored = set(ignored)

        if id(obj) in ignored:
            return self.recursive_value

        if isinstance(obj, dict):
            ignored.add(id(obj))
            seen.append(obj)

            clean_dict = {}
            for key, value in obj.items():
                if self._should_filter(key):
                    clean_dict[key] = self.filtered_value
                else:
                    clean_dict[key] = self.filter_string_values(
                        value, ignored, seen)

            return clean_dict

        return obj

    def default(self, obj):
        """
        Coerce values to strings if possible, otherwise replace with
        '[BADENCODING]'
        """
        try:
            if isinstance(obj, bytes):
                return str(obj, encoding='utf-8', errors='replace')
            else:
                return str(obj)

        except Exception:
            self.logger.exception('Could not add object to payload')
            return self.unencodeable_value

    def _sanitize(self, obj, trim_strings, ignored=None, seen=None):
        """
        Replace recursive values and trim strings longer than
        MAX_STRING_LENGTH
        """
        if not ignored:
            ignored = set()

        # Keep track of nested objects to avoid having references garbage
        # collected (which would cause id reuse and false positive recursion)
        if seen is None:
            seen = []

        if type(ignored) is list:
            ignored = set(ignored)

        if id(obj) in ignored:
            return self.recursive_value
        elif isinstance(obj, dict):
            ignored.add(id(obj))
            seen.append(obj)
            return self._sanitize_dict(obj, trim_strings, ignored, seen)
        elif isinstance(obj, (set, tuple, list)):
            ignored.add(id(obj))
            seen.append(obj)
            items = []
            for value in obj:
                items.append(
                        self._sanitize(value, trim_strings, ignored, seen))
            return items
        elif trim_strings and isinstance(obj, str):
            return obj[:MAX_STRING_LENGTH]
        else:
            return obj

    def _sanitize_dict_key_value(self, clean_dict, key, clean_value):
        """
        Safely sets the provided key on the dictionary by coercing the key
        to a string
        """
        if isinstance(key, bytes):
            try:
                key = str(key, encoding='utf-8', errors='replace')
                clean_dict[key] = clean_value
            except Exception:
                self.logger.exception(
                    'Could not add sanitize key for dictionary, '
                    'dropping value.')
        if isinstance(key, str):
            clean_dict[key] = clean_value
        else:
            try:
                clean_dict[str(key)] = clean_value
            except Exception:
                self.logger.exception(
                    'Could not add sanitize key for dictionary, '
                    'dropping value.')

    def _sanitize_dict(self, obj, trim_strings, ignored, seen):
        """
        Trim individual values in an object, applying filtering if the object
        is a FilterDict
        """
        if isinstance(obj, FilterDict):
            obj = self.filter_string_values(obj)

        clean_dict = {}
        for key, value in obj.items():

            clean_value = self._sanitize(value, trim_strings, ignored, seen)

            self._sanitize_dict_key_value(clean_dict, key, clean_value)

        return clean_dict

    def _should_filter(self, key):
        if isinstance(key, str):
            key_lower = key.lower()
            return any(f in key_lower for f in self.filters)

        if isinstance(key, bytes):
            key_lower = key.lower()
            return any(f in key_lower for f in self.bytes_filters)

        return False


class FilterDict(dict):
    """
    Object which will be filtered when encoded
    """
    pass


ContentType = Tuple[str, Optional[str], Optional[str], Optional[str]]


def parse_content_type(value: str) -> ContentType:
    """
    Generate a tuple of (type, subtype, suffix, parameters) from a type based
    on RFC 6838

    >>> parse_content_type("text/plain")
    ('text', 'plain', None, None)
    >>> parse_content_type("application/hal+json")
    ('application', 'hal', 'json', None)
    >>> parse_content_type("application/json;schema=\\"ftp://example.com/a\\"")
    ('application', 'json', None, 'schema="ftp://example.com/a"')
    """
    parameters = None  # type: Optional[str]
    if ';' in value:
        types, parameters = value.split(';', 1)
    else:
        types = value
    if '/' in types:
        maintype, subtype = types.split('/', 1)
        if '+' in subtype:
            subtype, suffix = subtype.split('+', 1)
            return (maintype, subtype, suffix, parameters)
        else:
            return (maintype, subtype, None, parameters)
    else:
        return (types, None, None, parameters)


def is_json_content_type(value: str) -> bool:
    """
    Check if a content type is JSON-parseable

    >>> is_json_content_type('text/plain')
    False
    >>> is_json_content_type('application/schema+json')
    True
    >>> is_json_content_type('application/json')
    True
    """
    type, subtype, suffix, _ = parse_content_type(value.lower())
    return type == 'application' and (subtype == 'json' or suffix == 'json')


def fully_qualified_class_name(obj):
    module = inspect.getmodule(obj)
    if module is not None and module.__name__ != "__main__":
        return module.__name__ + "." + obj.__class__.__name__
    else:
        return obj.__class__.__name__


def package_version(package_name):
    try:
        import pkg_resources
    except ImportError:
        return None
    else:
        try:
            return pkg_resources.get_distribution(package_name).version
        except pkg_resources.DistributionNotFound:
            return None


def _validate_setter(types, func, should_error=False):
    """
    Check that the first argument of a function is of a provided set of types
    before calling the body of the wrapped function, printing a runtime warning
    (or raising a TypeError) if the validation fails.
    """
    @wraps(func)
    def wrapper(obj, value):
        option_name = func.__name__
        if value is None or isinstance(value, types):
            func(obj, value)
        else:
            error_format = '{0} should be {1}, got {2}'
            actual = type(value).__name__
            requirement = ' or '.join([t.__name__ for t in types])
            message = error_format.format(option_name, requirement, actual)
            if should_error:
                raise TypeError(message)
            else:
                warnings.warn(message, RuntimeWarning)
    return wrapper


validate_str_setter = partial(_validate_setter, (str,))
validate_required_str_setter = partial(_validate_setter, (str,),
                                       should_error=True)
validate_bool_setter = partial(_validate_setter, (bool,))
validate_iterable_setter = partial(_validate_setter, (list, tuple))
validate_int_setter = partial(_validate_setter, (int,))


class ThreadContextVar:
    """
    A wrapper around thread-local variables to mimic the API of contextvars
    """
    LOCALS = None

    @classmethod
    def local_context(cls):
        if not ThreadContextVar.LOCALS:
            ThreadContextVar.LOCALS = threadlocal()
        return ThreadContextVar.LOCALS

    def __init__(self, name, **kwargs):
        self.name = name

        # Mimic the behaviour of ContextVar - if a default has been explicitly
        # passed then we will use it, otherwise don't set an initial value
        # This allows 'get' to know when to raise a LookupError
        if 'default' in kwargs:
            self.default = kwargs['default']
            # Make a deep copy so this thread starts with a fresh default
            self.set(copy.deepcopy(self.default))

    def get(self):
        local = ThreadContextVar.local_context()
        if hasattr(local, self.name):
            return getattr(local, self.name)

        if hasattr(self, 'default'):
            # Make a deep copy so that each thread starts with a fresh default
            result = copy.deepcopy(self.default)
            self.set(result)
            return result

        raise LookupError("No value for '{}'".format(self.name))

    def set(self, new_value):
        setattr(ThreadContextVar.local_context(), self.name, new_value)


def sanitize_url(url_to_sanitize: AnyStr) -> Optional[AnyStr]:
    try:
        parsed = urlparse(url_to_sanitize)

        sanitized_url = urlunsplit(
            # urlunsplit always requires 5 elements in this tuple
            (parsed.scheme, parsed.netloc, parsed.path, None, None)
        ).strip()
    except Exception:
        return None

    # If the sanitized url is empty then it did not have any of the components
    # we are interested in, so return None to indicate failure
    if not sanitized_url:
        return None

    return sanitized_url


# to_rfc3339: format a datetime instance to match to_rfc3339/iso8601 with
# milliseconds precision
# Python can do this natively from version 3.6, but we need to include a
# fallback implementation for Python 3.5
try:
    # this will raise if 'timespec' isn't supported
    datetime.utcnow().isoformat(timespec='milliseconds')  # type: ignore

    def to_rfc3339(dt: datetime) -> str:
        return dt.isoformat(timespec='milliseconds')  # type: ignore

except Exception:
    def _get_timezone_offset(dt: datetime) -> str:
        if dt.tzinfo is None:
            return ''

        utc_offset = dt.tzinfo.utcoffset(dt)

        if utc_offset is None:
            return ''

        sign = '+'

        if utc_offset.days < 0:
            sign = '-'
            utc_offset = -utc_offset

        hours_offset, minutes = divmod(utc_offset, timedelta(hours=1))
        minutes_offset, seconds = divmod(minutes, timedelta(minutes=1))

        return '{:s}{:02d}:{:02d}'.format(sign, hours_offset, minutes_offset)

    def to_rfc3339(dt: datetime) -> str:
        return '{:04d}-{:02d}-{:02d}T{:02d}:{:02d}:{:02d}.{:03d}{:s}'.format(
            dt.year,
            dt.month,
            dt.day,
            dt.hour,
            dt.minute,
            dt.second,
            int(dt.microsecond / 1000),
            _get_timezone_offset(dt)
        )



import builtins
import sys
import threading
import warnings

from datetime import datetime, timezone
from functools import wraps
from typing import Union, Tuple, Callable, Optional, List, Type, Dict, Any

from bugsnag.breadcrumbs import (
    Breadcrumb,
    BreadcrumbType,
    OnBreadcrumbCallback
)
from bugsnag.configuration import Configuration, RequestConfiguration
from bugsnag.event import Event
from bugsnag.handlers import BugsnagHandler
from bugsnag.sessiontracker import SessionTracker
from bugsnag.utils import to_rfc3339, fully_qualified_class_name as class_name

__all__ = ('Client',)


class Client:
    """
    A Bugsnag monitoring and reporting client.

    >>> client = Client(api_key='...')  # doctest: +SKIP
    """

    def __init__(self, configuration: Optional[Configuration] = None,
                 install_sys_hook=True, **kwargs):
        self.configuration = configuration or Configuration()  # type: Configuration  # noqa: E501
        self.session_tracker = SessionTracker(self.configuration)
        self.configuration.configure(**kwargs)

        if install_sys_hook:
            self.install_sys_hook()

    def capture(self,
                exceptions: Union[Tuple[Type, ...], Callable, None] = None,
                **options):
        """
        Run a block of code within the clients context.
        Any exception raised will be reported to bugsnag.

        >>> with client.capture():  # doctest: +SKIP
        ...     raise Exception('an exception passed to bugsnag then reraised')

        The context can optionally include specific types to capture.

        >>> with client.capture((TypeError,)):  # doctest: +SKIP
        ...     raise Exception('an exception which does get captured')

        Alternately, functions can be decorated to capture any
        exceptions thrown during execution and reraised.

        >>> @client.capture  # doctest: +SKIP
        ... def foo():
        ...     raise Exception('an exception passed to bugsnag then reraised')

        The decoration can optionally include specific types to capture.

        >>> @client.capture((TypeError,))  # doctest: +SKIP
        ... def foo():
        ...     raise Exception('an exception which does not get captured')
        """

        if callable(exceptions):
            return ClientContext(self, (Exception,))(exceptions)

        return ClientContext(self, exceptions, **options)

    def notify(self, exception: BaseException, asynchronous=None, **options):
        """
        Notify bugsnag of an exception.

        >>> client.notify(Exception('Example'))  # doctest: +SKIP
        """

        event = Event(exception, self.configuration,
                      RequestConfiguration.get_instance(), **options)

        self._leave_breadcrumb_for_event(event)
        self.deliver(event, asynchronous=asynchronous)

    def notify_exc_info(self, exc_type, exc_value, traceback,
                        asynchronous=None, **options):
        """
        Notify bugsnag of an exception via exc_info.

        >>> client.notify_exc_info(*sys.exc_info())  # doctest: +SKIP
        """

        exception = exc_value
        options['traceback'] = traceback
        event = Event(exception, self.configuration,
                      RequestConfiguration.get_instance(), **options)

        self._leave_breadcrumb_for_event(event)
        self.deliver(event, asynchronous=asynchronous)

    def excepthook(self, exc_type, exc_value, traceback):
        if self.configuration.auto_notify:
            self.notify_exc_info(
                exc_type, exc_value, traceback,
                severity='error',
                unhandled=True,
                severity_reason={
                    'type': 'unhandledException'
                })

    def install_sys_hook(self):
        self.sys_excepthook = sys.excepthook

        def excepthook(*exc_info):
            self.excepthook(*exc_info)

            if self.sys_excepthook:
                self.sys_excepthook(*exc_info)

        sys.excepthook = excepthook
        sys.excepthook.bugsnag_client = self

        if hasattr(threading, 'excepthook'):
            self.threading_excepthook = threading.excepthook

            def threadhook(args):
                self.excepthook(args[0], args[1], args[2])

                if self.threading_excepthook:
                    self.threading_excepthook(args)

            threading.excepthook = threadhook
            threading.excepthook.bugsnag_client = self

    def uninstall_sys_hook(self):
        client = getattr(sys.excepthook, 'bugsnag_client', None)

        if client is self:
            sys.excepthook = self.sys_excepthook
            self.sys_excepthook = None

        if hasattr(threading, 'excepthook'):
            client = getattr(threading.excepthook, 'bugsnag_client', None)
            if client is self:
                threading.excepthook = self.threading_excepthook
                self.threading_excepthook = None

    def deliver(self, event: Event,
                asynchronous: Optional[bool] = None):
        """
        Deliver the exception event to Bugsnag.
        """

        if not self.should_deliver(event):
            return

        def run_middleware():
            initial_severity = event.severity
            initial_reason = event.severity_reason.copy()

            def send_payload():
                if asynchronous is None:
                    options = {}
                else:
                    options = {'asynchronous': asynchronous}

                if event.api_key is None:
                    self.configuration.logger.warning(
                        "No API key configured, couldn't notify"
                    )

                    return

                if initial_severity != event.severity:
                    event.severity_reason = {
                        'type': 'userCallbackSetSeverity'
                    }
                else:
                    event.severity_reason = initial_reason
                payload = event._payload()
                try:
                    self.configuration.delivery.deliver(self.configuration,
                                                        payload, options)
                except Exception as e:
                    self.configuration.logger.exception(
                        'Notifying Bugsnag failed %s',
                        e
                    )

                # Trigger session delivery
                self.session_tracker.send_sessions()

            self.configuration.middleware.run(event, send_payload)

        self.configuration.internal_middleware.run(event, run_middleware)

    def should_deliver(self, event: Event) -> bool:
        # Return early if we shouldn't notify for current release stage
        if not self.configuration.should_notify():
            return False

        # Return early if we should ignore exceptions of this type
        if self.configuration.should_ignore(event.exception):
            return False

        return True

    def log_handler(self, extra_fields: List[str] = None) -> BugsnagHandler:
        return BugsnagHandler(client=self, extra_fields=extra_fields)

    @property
    def breadcrumbs(self) -> List[Breadcrumb]:
        return self.configuration.breadcrumbs

    def add_on_breadcrumb(self, on_breadcrumb: OnBreadcrumbCallback) -> None:
        self.configuration.add_on_breadcrumb(on_breadcrumb)

    def remove_on_breadcrumb(
        self,
        on_breadcrumb: OnBreadcrumbCallback
    ) -> None:
        self.configuration.remove_on_breadcrumb(on_breadcrumb)

    def leave_breadcrumb(
        self,
        message: str,
        metadata: Dict[str, Any] = {},
        type: BreadcrumbType = BreadcrumbType.MANUAL
    ) -> None:
        # Don't create breadcrumbs if the max_breadcrumbs is 0 as they would
        # be immediately discarded anyway
        if self.configuration.max_breadcrumbs == 0:
            return

        if not isinstance(type, BreadcrumbType):
            type = BreadcrumbType.MANUAL

        if not isinstance(metadata, dict):
            warning_message = 'breadcrumb metadata must be a dict, got {}'
            warnings.warn(
                warning_message.format(builtins.type(metadata).__name__),
                RuntimeWarning
            )

            metadata = {}

        timestamp = to_rfc3339(datetime.now(timezone.utc))
        breadcrumb = Breadcrumb(message, type, metadata, timestamp)

        for callback in self.configuration._on_breadcrumbs:
            try:
                should_continue = callback(breadcrumb)

                if should_continue is False:
                    self.configuration.logger.info(
                        'Breadcrumb not attached due to on_breadcrumb callback'
                    )

                    return
            except Exception:
                self.configuration.logger.exception(
                    'Exception raised in on_breadcrumb callback'
                )

        self.configuration._breadcrumbs.append(breadcrumb)

    def _auto_leave_breadcrumb(
        self,
        message: str,
        metadata: Dict[str, Any],
        type: BreadcrumbType
    ) -> None:
        if type in self.configuration.enabled_breadcrumb_types:
            self.leave_breadcrumb(message, metadata, type)

    def _leave_breadcrumb_for_event(self, event: Event) -> None:
        error_class = class_name(event.exception)

        self._auto_leave_breadcrumb(
            error_class,
            {
                'errorClass': error_class,
                'message': str(event.exception),
                'unhandled': event.unhandled,
                'severity': event.severity,
            },
            BreadcrumbType.ERROR
        )


class ClientContext:
    def __init__(self, client,
                 exception_types: Optional[Tuple[Type, ...]] = None,
                 **options):
        self.client = client
        self.options = options
        if 'severity' in options:
            options['severity_reason'] = dict(type='userContextSetSeverity')
        self.exception_types = exception_types or (Exception,)

    def __call__(self, function: Callable):
        @wraps(function)
        def decorate(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            except self.exception_types as e:
                self.client.notify(e, source_func=function, **self.options)
                raise

        return decorate

    def __enter__(self):
        pass

    def __exit__(self, *exc_info):
        if any(exc_info):
            if any(isinstance(exc_info[1], e) for e in self.exception_types):
                self.client.notify_exc_info(*exc_info, **self.options)

        return False



import warnings

running_uwsgi = False
threads_enabled = False

# From docs (https://uwsgi-docs.readthedocs.io/en/latest/PythonModule.html):
# > The uWSGI server automagically adds a uwsgi module into your Python
#   apps.
# > This is useful for configuring the uWSGI server, use its internal
#   functions and get statistics. Also useful for detecting whether you're
#   actually running under uWSGI; if you attempt to import uwsgi and
#   receive an ImportError you're not running under uWSGI.
try:
    # > uwsgi.opt
    # > The current configuration options, including any custom placeholders.
    from uwsgi import opt
    running_uwsgi = True
    threads_enabled = opt.get('enable-threads', False)
except ImportError:
    pass

__all__ = ('warn_if_running_uwsgi_without_threads',)


def warn_if_running_uwsgi_without_threads():
    if running_uwsgi and not threads_enabled:
        warnings.warn(('Bugsnag cannot run asynchronously under uWSGI' +
                       ' without enabling thread support. Please run uwsgi' +
                       ' with --enable-threads'), RuntimeWarning)



from bugsnag.configuration import Configuration, RequestConfiguration
from bugsnag.notification import Notification
from bugsnag.event import Event
from bugsnag.client import Client
from bugsnag.breadcrumbs import (
    BreadcrumbType,
    Breadcrumb,
    Breadcrumbs,
    OnBreadcrumbCallback
)
from bugsnag.legacy import (configuration, configure, configure_request,
                            add_metadata_tab, clear_request_config, notify,
                            auto_notify, before_notify, start_session,
                            auto_notify_exc_info, logger, leave_breadcrumb,
                            add_on_breadcrumb, remove_on_breadcrumb)

__all__ = ('Client', 'Event', 'Configuration', 'RequestConfiguration',
           'configuration', 'configure', 'configure_request',
           'add_metadata_tab', 'clear_request_config', 'notify',
           'auto_notify', 'before_notify', 'start_session',
           'auto_notify_exc_info', 'Notification', 'logger',
           'BreadcrumbType', 'Breadcrumb', 'Breadcrumbs',
           'OnBreadcrumbCallback', 'leave_breadcrumb', 'add_on_breadcrumb',
           'remove_on_breadcrumb')



import django
from django.conf import settings
from django.core.signals import request_started, got_request_exception

try:
    from django.core.urlresolvers import resolve, Resolver404
except ImportError:
    from django.urls import resolve, Resolver404

import bugsnag
from bugsnag.utils import is_json_content_type
import json


def add_django_request_to_notification(event):
    if not hasattr(event.request_config, "django_request"):
        return

    request = event.request_config.django_request
    event.request = request

    if event.context is None:
        try:
            route = resolve(request.path_info)
        except Resolver404:
            route = None

        if route and route.url_name:
            event.context = route.url_name
        elif route and route.view_name:
            event.context = route.view_name
        else:
            event.context = "%s %s" % (request.method, request.path_info)

    if hasattr(request, 'user'):
        if callable(request.user.is_authenticated):
            is_authenticated = request.user.is_authenticated()
        else:
            is_authenticated = request.user.is_authenticated
        if is_authenticated:
            try:
                name = request.user.get_full_name()
                email = getattr(request.user, 'email', None)
                username = str(request.user.get_username())
                event.set_user(id=username, email=email, name=name)
            except Exception:
                event.config.logger.exception('Could not get user data')
    else:
        event.set_user(id=request.META['REMOTE_ADDR'])

    if getattr(request, "session", None):
        event.add_tab("session", dict(request.session))
    request_tab = {
        'method': request.method,
        'path': request.path,
        'encoding': request.encoding,
        'GET': dict(request.GET),
        'POST': dict(request.POST),
        'url': request.build_absolute_uri(),
    }
    try:
        is_json = is_json_content_type(request.META.get('CONTENT_TYPE', ''))
        if is_json and request_tab["method"] == "POST":
            body = request.body.decode('utf-8', 'replace')
            request_tab["POST"] = json.loads(body)
    except Exception:
        pass

    event.add_tab("request", request_tab)
    if bugsnag.configure().send_environment:
        event.add_tab("environment", dict(request.META))


def configure():
    config = bugsnag.configure()

    # default to development if in DEBUG mode
    if getattr(settings, 'DEBUG'):
        config.configure(release_stage='development')

    # Import Bugsnag settings from settings.py
    django_bugsnag_settings = getattr(settings, 'BUGSNAG', {})
    config.configure(**django_bugsnag_settings)

    middleware = config.internal_middleware
    middleware.before_notify(add_django_request_to_notification)

    config.runtime_versions['django'] = django.__version__

    request_started.connect(__track_session)
    got_request_exception.connect(
        __handle_request_exception(config.logger),
        weak=False
    )

    return config


def __track_session(sender, **extra):
    if bugsnag.configuration.auto_capture_sessions:
        bugsnag.start_session()


def __handle_request_exception(logger):
    def inner(sender, **kwargs):
        request = kwargs.get('request', None)

        if request is not None:
            bugsnag.configure_request(django_request=request)

        try:
            bugsnag.auto_notify_exc_info(severity_reason={
                "type": "unhandledExceptionMiddleware",
                "attributes": {"framework": "Django"}
            })
        except Exception:
            logger.exception("Error in exception middleware")

    return inner



from enum import Enum
from typing import Any, List, Dict, Union, Optional

import bugsnag
from bugsnag.breadcrumbs import BreadcrumbType
from bugsnag.legacy import _auto_leave_breadcrumb
from bugsnag.utils import sanitize_url

__all__ = ('BugsnagMiddleware',)

SEVERITY_REASON = {
    "type": "unhandledExceptionMiddleware",
    "attributes": {
        "framework": "ASGI"
    }
}

DEFAULT_PORTS = {"http": 80, "https": 443, "ws": 80, "wss": 443}


def parse_host_header(headers: Dict[bytes, bytes]) -> Union[None, str]:
    if b'host' in headers:
        return headers[b'host'].decode('latin-1', 'ignore')
    return None


def parse_server_host(scheme: str, server) -> Union[None, str]:
    if len(server) == 2:
        hostname, port = server
        if port == DEFAULT_PORTS[scheme]:
            return hostname
        else:
            return '{}:{}'.format(hostname, port)
    return None


def parse_url(request: dict, server: List[Any]) -> str:
    scheme = request.get('scheme', 'http')
    query = request.get('query', None)
    path = request.get('path', '/')
    headers = request.get('headers', dict())
    host = parse_host_header(headers) or parse_server_host(scheme, server)
    url = ''
    if host is not None:
        url += '{}://{}'.format(scheme, host)
    url += path
    if query is not None and len(query) > 0:
        url += '?{}'.format(query.decode('utf-8', 'ignore'))

    return url


class RequestMetadata(Enum):
    http_method = ('httpMethod', ['method', 'http_method'])
    http_version = ('httpVersion', ['http_version'])
    path = ('path', ['path'])
    scheme = ('scheme', ['scheme'])
    type = ('type', ['type'])
    query = ('query', ['query_string'])

    def __init__(self, metadata_key: str, scope_keys: List[str]):
        self.metadata_key = metadata_key
        self.scope_keys = scope_keys


class BugsnagMiddleware:
    """
    Sends unhandled exceptions to Bugsnag which happen while processing
    requests.

    >>> async def app(scope, receive, send):
    ...     await send({
    ...         "type": "http.request",
    ...         "body": b"Hello World",
    ...         "more_body": False,
    ...     })
    >>> app = BugsnagMiddleware(app)
    """
    def __init__(self, app):
        self.app = app
        stack = bugsnag.configure().internal_middleware

        def add_request_info(event):
            if not hasattr(event.request_config, 'asgi_scope'):
                return

            scope = event.request_config.asgi_scope
            request = dict()
            server = []

            client = scope.get('client')
            if client is not None and len(client) > 0:
                request['clientIp'] = scope['client'][0]

            if 'server' in scope and type(scope['server']) in [list, tuple]:
                server = scope['server']
            if 'headers' in scope:
                request['headers'] = dict(
                        [i for i in scope['headers'] if len(i) == 2])
            for prop in RequestMetadata:
                for item in prop.scope_keys:
                    if item in scope:
                        request[prop.metadata_key] = scope[item]
                        break

            request['url'] = parse_url(request, server)

            event.add_tab("request", request)
            if bugsnag.configure().send_environment:
                event.add_tab("environment", scope)

        stack.before_notify(add_request_info)

    async def __call__(self, scope, receive, send):
        bugsnag.configure()._breadcrumbs.create_copy_for_context()
        bugsnag.configure_request(asgi_scope=scope)
        try:
            if bugsnag.configuration.auto_capture_sessions:
                bugsnag.start_session()

            # only HTTP and Websocket requests have headers
            if scope['type'] in ('http', 'websocket'):
                _auto_leave_breadcrumb(
                    '{} request'.format(scope['type']),
                    _get_breadcrumb_metadata(scope),
                    BreadcrumbType.NAVIGATION
                )

            await self.app(scope, receive, send)
        except Exception as e:
            bugsnag.auto_notify(e, severity_reason=SEVERITY_REASON)
            raise


def _get_breadcrumb_metadata(scope) -> Dict[str, str]:
    metadata = {'to': scope['path']}
    referer = _get_referer_header(scope)

    if referer:
        metadata['from'] = sanitize_url(referer)

    return metadata


def _get_referer_header(scope) -> Optional[str]:
    for key, value in scope['headers']:
        if key == b'referer':
            return value.decode('latin-1', 'ignore')

    return None



import tornado
from tornado.web import RequestHandler, HTTPError
from tornado.wsgi import WSGIContainer
from typing import Dict, Any
from urllib.parse import parse_qs
from bugsnag.breadcrumbs import BreadcrumbType
from bugsnag.utils import is_json_content_type, sanitize_url
from bugsnag.legacy import _auto_leave_breadcrumb
import bugsnag
import json


class BugsnagRequestHandler(RequestHandler):
    def add_tornado_request_to_notification(self, event: bugsnag.Event):
        if not hasattr(self, "request"):
            return

        event.request = self.request
        request_tab = {
            'method': self.request.method,
            'path': self.request.path,
            'GET': parse_qs(self.request.query),
            'POST': {},
            'url': self.request.full_url(),
        }  # type: Dict[str, Any]
        try:
            if (len(self.request.body) > 0):
                headers = self.request.headers
                body = self.request.body.decode('utf-8', 'replace')
                is_json = is_json_content_type(headers.get('Content-Type', ''))
                if is_json and request_tab["method"] == "POST":
                    request_tab["POST"] = json.loads(body)
                else:
                    request_tab["POST"] = self.request.body_arguments
        except Exception:
            pass

        event.add_tab("request", request_tab)

        if bugsnag.configure().send_environment:
            env = WSGIContainer.environ(self.request)
            event.add_tab("environment", env)

    def _handle_request_exception(self, exc: BaseException):
        options = {
            "user": {"id": self.request.remote_ip},
            "context": self._get_context(),
            "request": {
                "url": self.request.full_url(),
                "method": self.request.method,
                "arguments": self.request.arguments,
            },
            "severity_reason": {
                "type": "unhandledExceptionMiddleware",
                "attributes": {
                    "framework": "Tornado"
                }
            }
        }

        # Notify bugsnag, unless it's an HTTPError that we specifically want
        # to ignore
        should_notify_bugsnag = True
        if isinstance(exc, HTTPError):
            ignore_status_codes = self.bugsnag_ignore_status_codes()
            if ignore_status_codes and exc.status_code in ignore_status_codes:
                should_notify_bugsnag = False

        if should_notify_bugsnag:
            bugsnag.auto_notify(exc, **options)

        # Call the parent handler
        RequestHandler._handle_request_exception(self, exc)  # type: ignore

    def prepare(self):
        middleware = bugsnag.configure().internal_middleware
        bugsnag.configure().runtime_versions['tornado'] = tornado.version
        middleware.before_notify(self.add_tornado_request_to_notification)
        bugsnag.configure()._breadcrumbs.create_copy_for_context()

        _auto_leave_breadcrumb(
            'http request',
            self._get_breadcrumb_metadata(),
            BreadcrumbType.NAVIGATION
        )

        if bugsnag.configuration.auto_capture_sessions:
            bugsnag.start_session()

    def _get_breadcrumb_metadata(self) -> Dict[str, str]:
        if not hasattr(self, 'request'):
            return {}

        metadata = {'to': self.request.path}

        if 'Referer' in self.request.headers:
            metadata['from'] = sanitize_url(self.request.headers['Referer'])

        return metadata

    def _get_context(self):
        return "%s %s" % (self.request.method, self.request.uri.split("?")[0])

    def bugsnag_ignore_status_codes(self):
        # Subclasses can override to add or remove codes
        return range(400, 500)
